using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;

// VAMCUSTOMBUTTONS v1.0
//
// Customization system for VAM's UIButton and UIText

namespace VAMCUIPlugin
{
    public class VAMCUI : MVRScript
    {
        public static string PLUGIN_PATH;
        public static string ASSETS_PATH;
        public static string ASSETS_BUNDLE_PATH;
		
		// Game Objects and components
		private GameObject VAMCUIGO;	
		private GameObject VAMCUIBgGO;
		private Image VAMCUIBgImg;
		
		private GameObject VAMCUITxtGO;
		private Text VAMCUITxtText;
		
		public FreeControllerV3 CUIAtomControl;
		
		public JSONStorableStringChooser CUITxtFontChoice;
		public JSONStorableStringChooser CUITxtAlignmentChoice;
		public JSONStorableStringChooser CUITxtBgChoice;
		public JSONStorableStringChooser CUITxtMatChoice;
		
		public JSONStorableStringChooser setSubtitlesFont;
		public JSONStorableStringChooser setSubtitlesAlignment;
		
		public JSONStorableColor MaterialDetailsColor;
		public JSONStorableFloat MaterialDetailsSpeedX;
		public JSONStorableFloat MaterialDetailsSpeedY;
		public JSONStorableFloat MaterialSubdetailsOpacity;
		public JSONStorableFloat MaterialDetailsRefractionIntensity;
		
		public JSONStorableBool CUIRotateToCamera;
		public JSONStorableBool CUIRotateConstrainToY;
					
		protected List<string> alignmentChoices;
		protected List<string> fontChoices;
		protected List<string> bgChoices;
		protected List<string> matChoices;
		
		protected Dictionary<string, string> FontList;
		protected Dictionary<string, Font> FontAssets;
		protected Dictionary<string, string> BgList;
		protected Dictionary<string, Sprite> BgAssets;
		protected Dictionary<string, string> MaterialList;
		protected Dictionary<string, Material> MaterialAssets;
		
		protected Sprite CUIOriginalSpr;
		protected Material CUIOriginalMat;
		protected Vector2 CUITxtOriginalOffsetMin;
		protected Vector2 CUITxtOriginalOffsetMax;
		protected TextAnchor CUITxtOriginalTextAnchor;
			
		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "UIButton" && containingAtom.type != "UIText")
                {
                    SuperController.LogError("Please add VAMCUI on a UIButton Atom");
                    return;
                }
				
				PLUGIN_PATH = GetPluginPath(this);
				ASSETS_PATH = PLUGIN_PATH + "/assets";
				ASSETS_BUNDLE_PATH = ASSETS_PATH + "/vamcui.assetbundle";
				
				// Alignments
				alignmentChoices = new List<string>();
				alignmentChoices.Add("Top Left");
				alignmentChoices.Add("Top Center");
				alignmentChoices.Add("Top Right");
				alignmentChoices.Add("Middle Left");
				alignmentChoices.Add("Middle Center");
				alignmentChoices.Add("Middle Right");
				alignmentChoices.Add("Bottom Left");
				alignmentChoices.Add("Bottom Center");
				alignmentChoices.Add("Bottom Right");
				
				// Fonts paths in the bundle. I'm lazy, and I don't think I do have benefits from loading a json file just for simple paths
				FontList = new Dictionary<string, string>();
				FontList.Add("Amalia","Assets/VAMFonts/amalia.ttf");
				FontList.Add("AnimeAceBB","Assets/VAMFonts/animeacebb_reg.ttf");
				FontList.Add("Astera","Assets/VAMFonts/astera_v2.ttf");
				FontList.Add("Bebas","Assets/VAMFonts/bebas_regular.ttf");
				FontList.Add("BirdsOfParadise","Assets/VAMFonts/birds_of_paradise.ttf");
				FontList.Add("GlitchInside","Assets/VAMFonts/glitch_inside.otf");
				FontList.Add("GosmickSans","Assets/VAMFonts/gosmick_sans.ttf");
				FontList.Add("HomerSimpson","Assets/VAMFonts/homer_simpson_revised.ttf");
				FontList.Add("Kiona","Assets/VAMFonts/kiona_regular.ttf");
				FontList.Add("Oswald","Assets/VAMFonts/oswald.ttf");
				FontList.Add("PlanetKosmos","Assets/VAMFonts/planet_kosmos.ttf");
				FontList.Add("Poppins","Assets/VAMFonts/poppins.ttf");
				FontList.Add("SpaceAge","Assets/VAMFonts/space_age.ttf");
							
				// Fonts Assets, i'm creating Arial by default because this font rocks for subtitles, and if the bundle fails, we have at least one font
				FontAssets = new Dictionary<string, Font>();
				FontAssets.Add("Arial", Resources.GetBuiltinResource(typeof(Font), "Arial.ttf") as Font);
				
				// Creating font choices based on the dict
				fontChoices = new List<string>();
				fontChoices.Add("Arial"); // Adding Arial by default
				foreach( KeyValuePair<string, string> fontKvp in FontList )
				{
					// Kinda hacky, because I cannot find a way to update a popup after creating it
					// I'm gonna assume that there is never gonna be any problem to load the bundle :x
					fontChoices.Add(fontKvp.Key);
				}
				
				
				BgList = new Dictionary<string, string>();
				BgList.Add("Basic01","Assets/VAMTextures/CUI/cui_bg_custom_basic_01.png");
				BgList.Add("Basic02","Assets/VAMTextures/CUI/cui_bg_custom_basic_02.png");
				BgList.Add("Modern01","Assets/VAMTextures/CUI/cui_bg_custom_modern_01.png");
				BgList.Add("Modern02","Assets/VAMTextures/CUI/cui_bg_custom_modern_02.png");
				BgList.Add("Modern03","Assets/VAMTextures/CUI/cui_bg_custom_modern_03.png");
				BgList.Add("Modern04","Assets/VAMTextures/CUI/cui_bg_custom_modern_04.png");
				BgList.Add("Modern05","Assets/VAMTextures/CUI/cui_bg_custom_modern_05.png");
				BgList.Add("Fantasy01","Assets/VAMTextures/CUI/cui_bg_custom_fantasy_01.png");
				
				BgAssets = new Dictionary<string, Sprite>();
				
				// Creating Bg choices based on the dict
				bgChoices = new List<string>();
				foreach( KeyValuePair<string, string> bgKvp in BgList )
				{
					bgChoices.Add(bgKvp.Key);
				}
				
				MaterialList = new Dictionary<string, string>();
				MaterialList.Add("Material01","Assets/VAMMaterials/UI/CUI/cui_mat_01.mat");
				MaterialList.Add("Material02","Assets/VAMMaterials/UI/CUI/cui_mat_02.mat");
				MaterialList.Add("Material03","Assets/VAMMaterials/UI/CUI/cui_mat_03.mat");
				MaterialList.Add("Material04","Assets/VAMMaterials/UI/CUI/cui_mat_04.mat");
				MaterialList.Add("Material05","Assets/VAMMaterials/UI/CUI/cui_mat_05.mat");
				MaterialList.Add("Material06","Assets/VAMMaterials/UI/CUI/cui_mat_06.mat");
				
				MaterialAssets = new Dictionary<string, Material>();
				
				// Creating Mat choices based on the dict
				matChoices = new List<string>();
				matChoices.Add("None");
				foreach( KeyValuePair<string, string> matKvp in MaterialList )
				{
					matChoices.Add(matKvp.Key);
				}
				
				// Loading the bundle
				Request request = new AssetLoader.AssetBundleFromFileRequest {path = ASSETS_BUNDLE_PATH, callback = OnAssetBundleLoaded};
				AssetLoader.QueueLoadAssetBundleFromFile(request);
				
				// **************************************************
				// ** LEFT : HELP
				// **************************************************

				JSONStorableString helpText = new JSONStorableString("Help",
					"<color=#000><size=35><b>Custom UI Help</b></size></color>\n\n" + 
					"<color=#333>" +
					"<b>Font :</b> The type of font for the text. You can change the color and opacity in the \"Button\" or \"Text\" tab. <b>WARNING : </b> some fonts may disappear because of the size selected in the button or text tab. If so, just go in the button tab and lower the size of the text or make the button bigger.\n\n" +
					"<b>Text alignement :</b> How the text is aligned inside the UI\n\n" +
					"<b>Background :</b> Background image for the UI. You can change the color and opacity in the \"Button\" or \"Text\" tab.\n\n" +
					"<b>Rotation follows the camera :</b> If enabled, the UI will follow the camera.\n\n" +
					"<b>Rotation constrained to Y axis :</b> Only works if follow is enabled. Constrain the UI to only rotate on the vertical axis.\n\n\n" +
					"<b>Material :</b> Adds a material effect to the background with animation and refraction.\n\n" +
					"<b>Details color :</b> Sets the color of the animated parts of the material.\n\n" +
					"<b>Details scroll speed X :</b> Sets the speed of the animation on the horizontal axis.\n\n" +
					"<b>Details scroll speed Y :</b> Sets the speed of the animation on the vertical axis.\n\n" +
					"<b>Subdetails opacity :</b> Sets the visibility of an additionnal layer of animated details (by default it is not visible).\n\n" +
					"<b>Details refraction intensity :</b> Changes the refraction intensity of the background and details (this is mainly visible on transparent UI).\n\n" +
					"</color>"
				);
				UIDynamic helpTextfield = CreateTextField(helpText, false);
				helpTextfield.height = 1150.0f;
				
				// **************************************************
				// ** RIGHT : Options
				// **************************************************										
				// TITLE GLOBAL
				createStaticDescriptionText("Global options","<color=#000><size=35><b>BASIC OPTIONS</b></size></color>",true,40);
				
				CUITxtFontChoice = new JSONStorableStringChooser("Font", fontChoices, "Arial", "Font", CUITxtFontCallback);
				UIDynamicPopup SFCUdp = CreatePopup(CUITxtFontChoice, true);
				SFCUdp.labelWidth = 150f;
				
				CUITxtAlignmentChoice = new JSONStorableStringChooser("Text alignment", alignmentChoices, "Middle Center", "Text alignment", CUITxtAlignementCallback);
				UIDynamicPopup SACUdp = CreatePopup(CUITxtAlignmentChoice, true);
				SACUdp.labelWidth = 300f;
				
				CUITxtBgChoice = new JSONStorableStringChooser("Background", bgChoices, "Modern01", "Background", BackgroundCallback);
				UIDynamicPopup BTBCdp = CreatePopup(CUITxtBgChoice, true);
				BTBCdp.labelWidth = 200f;
				
				CUIRotateToCamera = new JSONStorableBool("Rotation follows the camera", false);
                CUIRotateToCamera.storeType = JSONStorableParam.StoreType.Full;
                RegisterBool(CUIRotateToCamera);
                CreateToggle(CUIRotateToCamera, true);
				
				CUIRotateConstrainToY = new JSONStorableBool("Rotation constrained to Y axis", false);
                CUIRotateConstrainToY.storeType = JSONStorableParam.StoreType.Full;
                RegisterBool(CUIRotateConstrainToY);
                CreateToggle(CUIRotateConstrainToY, true);
				
				CreateSpacer(true);
				createStaticDescriptionText("Global options","<color=#000><size=35><b>ADVANCED OPTIONS</b></size></color>",true,40);
				
				CUITxtMatChoice = new JSONStorableStringChooser("Material", matChoices, "Material01", "Material", MaterialCallback);
				UIDynamicPopup BTMCdp = CreatePopup(CUITxtMatChoice, true);
				BTMCdp.labelWidth = 200f;
				
				HSVColor hsvc = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
				MaterialDetailsColor = new JSONStorableColor("Details color", hsvc, MaterialDetailsColorCallback);
				RegisterColor(MaterialDetailsColor);
				CreateColorPicker(MaterialDetailsColor, true);
			
				MaterialDetailsSpeedX = new JSONStorableFloat("Details scroll speed X", 0.05f, MaterialDetailsSpeedXCallback, -2f, 2f);
				RegisterFloat(MaterialDetailsSpeedX);
				CreateSlider(MaterialDetailsSpeedX, true);
				
				MaterialDetailsSpeedY = new JSONStorableFloat("Details scroll speed Y", 0.05f, MaterialDetailsSpeedYCallback, -2f, 2f);
				RegisterFloat(MaterialDetailsSpeedY);
				CreateSlider(MaterialDetailsSpeedY, true);		

				MaterialSubdetailsOpacity = new JSONStorableFloat("Subdetails opacity", 0f, MaterialSubdetailsOpacityCallback, 0f, 1f);
				RegisterFloat(MaterialSubdetailsOpacity);
				CreateSlider(MaterialSubdetailsOpacity, true);	

				MaterialDetailsRefractionIntensity = new JSONStorableFloat("Details Refraction Intensity", 0.5f, MaterialDetailsRefractionIntensityCallback, 0f, 1f);
				RegisterFloat(MaterialDetailsRefractionIntensity);
				CreateSlider(MaterialDetailsRefractionIntensity, true);
				
				// **********************************************************************************************************************************
				// Registering variables, a bit strange to disconnect them from the initial creation, but allows me to order them in the action list
				// **********************************************************************************************************************************
				
				RegisterStringChooser(CUITxtFontChoice);
				RegisterStringChooser(CUITxtAlignmentChoice);
				RegisterStringChooser(CUITxtBgChoice);
				RegisterStringChooser(CUITxtMatChoice);

	
			}
            catch(Exception e)
            {
                SuperController.LogError("VAMCUI - Exception caught: " + e);
            }
		}
		
		void Start()
        {
			try
            {
				InitButtonObjects();
				
				CUITxtFontCallback(CUITxtFontChoice.val);
				CUITxtAlignementCallback(CUITxtAlignmentChoice.val);
				BackgroundCallback(CUITxtBgChoice.val);
				MaterialCallback(CUITxtMatChoice.val);
			}
            catch(Exception e)
            {
                SuperController.LogError("VAMCUI - Exception caught: " + e);
            }
        }
		
		void Update() {
			// Handling forced rotation
			// Based on Spacedog code, thank you buddy!
			if( CUIRotateToCamera.val == true ) {
				Vector3 lookPos = CameraTarget.centerTarget.transform.position - CUIAtomControl.transform.position;
				Quaternion lookRot = Quaternion.LookRotation(lookPos, Vector3.up);
				float eulerX = 0;
				if (CUIRotateConstrainToY.val == false)
				{
					eulerX = lookRot.eulerAngles.x;
				}
				float eulerY = lookRot.eulerAngles.y;
				Quaternion rotation = Quaternion.Euler (eulerX, eulerY, 0);
				CUIAtomControl.transform.rotation = rotation;
			}
		}
		
		// **************************
		// Functions
		// **************************
		

		
		protected void ChangeButtonFont( string fontVal ) {
			if( VAMCUITxtText != null && FontAssets.ContainsKey(fontVal) ) {
				VAMCUITxtText.font = FontAssets[fontVal];
			}
		}
		
		protected void ChangeButtonAlignment( string alignmentVal ) {
			if( VAMCUITxtText != null ) {
				if( alignmentVal == "Top Left" ) {
					VAMCUITxtText.alignment = TextAnchor.UpperLeft;
				} else if( alignmentVal == "Top Center" ) {
					VAMCUITxtText.alignment = TextAnchor.UpperCenter;
				} else if( alignmentVal == "Top Right" ) {
					VAMCUITxtText.alignment = TextAnchor.UpperRight;
				} else if( alignmentVal == "Middle Left" ) {
					VAMCUITxtText.alignment = TextAnchor.MiddleLeft;
				} else if( alignmentVal == "Middle Center" ) {
					VAMCUITxtText.alignment = TextAnchor.MiddleCenter;
				} else if( alignmentVal == "Middle Right" ) {
					VAMCUITxtText.alignment = TextAnchor.MiddleRight;
				} else if( alignmentVal == "Bottom Left" ) {
					VAMCUITxtText.alignment = TextAnchor.LowerLeft;
				} else if( alignmentVal == "Bottom Center" ) {
					VAMCUITxtText.alignment = TextAnchor.LowerCenter;
				} else {					
					VAMCUITxtText.alignment = TextAnchor.LowerRight;
				}
			}
		}
		
		protected void ChangeButtonBackground( string bgVal ) {
			if( VAMCUIBgImg != null && BgAssets.ContainsKey(bgVal) ) {
				VAMCUIBgImg.sprite = BgAssets[bgVal];
			}
		}
		
		protected void ChangeButtonMaterial( string matVal ) {
			if( VAMCUIBgImg != null  ) {
				if( matVal != "None" && MaterialAssets.ContainsKey(matVal) ) {
					VAMCUIBgImg.material = MaterialAssets[matVal];
				} else if( matVal == "None" ) {
					VAMCUIBgImg.material = CUIOriginalMat;
				}
			}
		}
		
		protected void SetMaterialOtherParams() {
			ChangeButtonAlignment( CUITxtAlignmentChoice.val );
			MaterialDetailsColorCallback( MaterialDetailsColor );
			MaterialDetailsSpeedXCallback( MaterialDetailsSpeedX );
			MaterialDetailsSpeedYCallback( MaterialDetailsSpeedY );
			MaterialSubdetailsOpacityCallback( MaterialSubdetailsOpacity );
			MaterialDetailsRefractionIntensityCallback( MaterialDetailsRefractionIntensity );
		}
		
		protected void InitButtonObjects() {
			try
            {
				// ******************************
				// Getting button components
				// ******************************
				CUIAtomControl = containingAtom.GetStorableByID("control") as FreeControllerV3;

				VAMCUIGO = getCUICanvas( containingAtom.reParentObject ).gameObject;
				
				Transform VAMCUIBgTR = getCUIBg( VAMCUIGO.transform );
				Transform VAMCUITxtTR = getCUIText( VAMCUIGO.transform );		
				
				VAMCUIBgGO = VAMCUIBgTR.gameObject;
				VAMCUITxtGO = VAMCUITxtTR.gameObject;
				
				VAMCUIBgImg = VAMCUIBgGO.GetComponent<Image>();
							
				VAMCUITxtText = VAMCUITxtGO.GetComponent<Text>();
				
				RectTransform VAMCUITxtRecTr = VAMCUITxtGO.GetComponent<RectTransform>();										
				CUIOriginalMat = VAMCUIBgImg.material;
				CUIOriginalSpr = VAMCUIBgImg.sprite;			
				CUITxtOriginalOffsetMin = VAMCUITxtRecTr.offsetMin;
				CUITxtOriginalOffsetMax = VAMCUITxtRecTr.offsetMax;
				CUITxtOriginalTextAnchor = VAMCUITxtText.alignment;
				
				VAMCUITxtRecTr.offsetMin = new Vector2(30.0f, 20.0f);
				VAMCUITxtRecTr.offsetMax = new Vector2(-30.0f, -20.0f);
			}
            catch(Exception e)
            {
                SuperController.LogError("VAMCUI - Exception caught while initializing: " + e);
            }
		}
		
		// **************************
		// Callbacks & similar
		// **************************
		private void OnAssetBundleLoaded(Request aRequest) {
			// Loading font assets
			foreach( KeyValuePair<string, string> fontKvp in FontList )
			{
				Font fnt = aRequest.assetBundle.LoadAsset<Font>(fontKvp.Value);
				if( fnt != null ) {
					// Adding font asset if successfuly loaded in the assets
					FontAssets.Add(fontKvp.Key, fnt);
					if( CUITxtFontChoice.val == fontKvp.Key ) {
						ChangeButtonFont(fontKvp.Key);
					}
				}
			}
						
			// Loading bg assets
			foreach( KeyValuePair<string, string> bgKvp in BgList )
			{
				Sprite spr = aRequest.assetBundle.LoadAsset<Sprite>(bgKvp.Value);
				if( spr != null ) {
					// Adding font asset if successfuly loaded in the assets
					BgAssets.Add(bgKvp.Key, spr);
					if( CUITxtBgChoice.val == bgKvp.Key ) {
						ChangeButtonBackground(bgKvp.Key);
					}
				}
			}
					
			// Loading material assets
			foreach( KeyValuePair<string, string> matKvp in MaterialList )
			{
				Material mat = aRequest.assetBundle.LoadAsset<Material>(matKvp.Value);
				if( mat != null ) {
					// Adding font asset if successfuly loaded in the assets
					MaterialAssets.Add(matKvp.Key, new Material(mat));
					if( CUITxtMatChoice.val == matKvp.Key ) {
						ChangeButtonMaterial(matKvp.Key);
					}
				}
			}
			
			// Settings all the other params
			SetMaterialOtherParams();
		}
				
		protected void CUITxtFontCallback(string fontchoice) {
			ChangeButtonFont( fontchoice );
		}
		
		protected void CUITxtAlignementCallback(string alignementchoice) {
			ChangeButtonAlignment( alignementchoice );
		}
		
		protected void BackgroundCallback(string backgroundchoice) {
			ChangeButtonBackground( backgroundchoice );
		}
		
		protected void MaterialCallback(string materialchoice) {
			ChangeButtonMaterial( materialchoice );
			
			// Setting the others params
			SetMaterialOtherParams();
		}
		
		protected void MaterialDetailsColorCallback(JSONStorableColor color) {
			if( VAMCUIBgImg != null ) {
				Color newColor = new Color(color.colorPicker.currentColor.r, color.colorPicker.currentColor.g, color.colorPicker.currentColor.b, 1.0f);
				VAMCUIBgImg.material.SetColor("_DetailsColor",newColor);
			}
		}
		
		protected void MaterialDetailsSpeedXCallback(JSONStorableFloat speedX) {
			if( VAMCUIBgImg != null ) {
				VAMCUIBgImg.material.SetFloat("_DetailsSpeedX", speedX.val);
			}
		}
		
		protected void MaterialDetailsSpeedYCallback(JSONStorableFloat speedY) {
			if( VAMCUIBgImg != null ) {
				VAMCUIBgImg.material.SetFloat("_DetailsSpeedY", speedY.val);
			}
		}
		
		protected void MaterialSubdetailsOpacityCallback(JSONStorableFloat opacity) {
			if( VAMCUIBgImg != null ) {
				VAMCUIBgImg.material.SetFloat("_SubdetailsOpacity", opacity.val);
			}
		}
		
		protected void MaterialDetailsRefractionIntensityCallback(JSONStorableFloat intensity) {
			if( VAMCUIBgImg != null ) {
				VAMCUIBgImg.material.SetFloat("_RefractionIntensity", intensity.val);
			}
		}
			
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private static void disableScrollOnText(UIDynamicTextField target) {
			ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
			if( targetSR != null ) {
				targetSR.horizontal = false;
				targetSR.vertical = false;
			}
		}
		
		private void createStaticDescriptionText( string DescTitle, string DescText, bool rightSide, int fieldHeight ) {
			JSONStorableString staticDescString = new JSONStorableString(DescTitle,DescText);
			UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
			staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
			sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
			staticDescStringField.height = fieldHeight;
			disableScrollOnText(staticDescStringField);
		}
		
        // Recursive function searching for a transform containing the string set in objectNameSearchString
        // It is used the first time for the initialization of the script
        private Transform getCUICanvas( Transform parent ) {
            foreach( Transform child in parent )
            {
				if( child.name == "Canvas" ) {
					return child;
				} else {
					Transform childSearch = getCUICanvas(child);
					if (childSearch != null) return childSearch;
				}
            }
        
            return null;
        }
		
        // Recursive function searching for the custom button background
        private Transform getCUIBg( Transform parent ) {
			String bgName = "Button";
			if (containingAtom.type == "UIText") {
				bgName = "Panel";
			}
            foreach( Transform child in parent )
            {
				if( child.name == bgName ) {
					return child;
				} else {
					Transform childSearch = getCUIBg(child);
					if (childSearch != null) return childSearch;
				}
            }
        
            return null;
        }
		
        // Recursive function searching for the custom button text
        private Transform getCUIText( Transform parent ) {
            foreach( Transform child in parent )
            {
				if( child.name == "Text" ) {
					return child;
				} else {
					Transform childSearch = getCUIText(child);
					if (childSearch != null) return childSearch;
				}
            }
        
            return null;
        }
		
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy() {
			// Cleaning modified native VAM assets
			if( VAMCUIBgImg ) {
				VAMCUIBgImg.material = CUIOriginalMat;
				VAMCUIBgImg.sprite = CUIOriginalSpr;
			}
			
			if( VAMCUITxtGO ) {
				RectTransform VAMCUITxtRecTr = VAMCUITxtGO.GetComponent<RectTransform>();										
				
				VAMCUITxtRecTr.offsetMin = CUITxtOriginalOffsetMin;
				VAMCUITxtRecTr.offsetMax = CUITxtOriginalOffsetMax;
				VAMCUITxtText.alignment = CUITxtOriginalTextAnchor;		
			}
			
			// Font bundle unload
			AssetLoader.DoneWithAssetBundleFromFile(ASSETS_BUNDLE_PATH);
		}
		
		// ***********************************************************
		// EXTERNAL TOOL - Thank you great coders for your content!
		// ***********************************************************
		
		// *********** MacGruber_Utils.cs START *********************
		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
				
		// Get path prefix of the package that contains our plugin.
		public static string GetPackagePath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/");
			if (idx >= 0)
				return filename.Substring(0, idx+2);
			else
				return string.Empty;
		}
				
		// Check if our plugin is running from inside a package
		public static bool IsInPackage(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
			string filename = self.manager.GetJSON()["plugins"][id].Value;
			return filename.IndexOf(":/") >= 0;
		}
		// *********** MacGruber_Utils.cs END *********************
	}
}